%module hashcash
%{
#include "sha1.h"
#define GROUP_SIZE ULONG_MAX
#define GROUP_DIGITS 8
#define GFORMAT "%08x"

void random_start( word32*, word32*, word32*, word32* );
word32 find_collision( char *day, char* name, int bits, 
		       char* collision, word32 tries, char* counter );
typedef struct _SHA1Obj {
	SHA1_ctx ctx ;
} sha1 ;

%}
%typemap(python,in) PyObject * {
        $target = $source;
}
%typemap(python,out) PyObject * {
        $target = $source;
}

typedef struct {
// Creates an object which computes the SHA message digest.
	~sha1();
	%addmethods {
	sha1() {
		sha1 *s = malloc(sizeof(sha1));
		SHA1_init(&(s->ctx)) ;
		return s;
	} ;
	PyObject *update(PyObject *s) {
		unsigned char *data ;
		word32 l;
		data = PyString_AsString(s) ;
		l = PyString_Size(s) ;
		SHA1_update(&(self->ctx),(void *)data,l);
		Py_INCREF(Py_None);
		return Py_None;
	} ;
	// Add a new string to the message digest.
	// Returns None.
	PyObject *digest() {
		PyObject *o ;
		static unsigned char md[20];
		SHA1_final(&(self->ctx));
		SHA1_digest(&(self->ctx),md);
		o = PyString_FromStringAndSize(md,20);
		return o;
	} ;
	// Compute the final message digest.
	}
} sha1 ;

%wrapper %{

PyObject *sha1_string(PyObject *args) {
	unsigned char *data;
	static unsigned char md[20] ;
	unsigned long l;
	SHA1_ctx ctx;
	PyObject *o;

	data = PyString_AsString(args) ;
	l = PyString_Size(args) ;
	SHA1_init(&ctx);
	SHA1_update(&ctx,data,l);
	SHA1_final(&ctx);
	SHA1_digest(&ctx,md);
	o = PyString_FromStringAndSize(md,20);
	return o;
}
%}

PyObject *sha1_string(PyObject *s);
// Compute SHA1 for the string, return the message digest string.

int still_valid( char* check, int validity_period );
// Returns the number of days the hashcash (check)
// will still be valid, based on validity_period.
int count_collision_bits( char* name, char* check );
// Return the number of bits of collision based on the
// name and the hashcash (check).
char *mint(char * today, char* name, int bits );
// Returns a new piece of hash cash. today is a five digit, zero-padded
// string representing number of days since epoch
// (today="%05d"%(time.time()/3600/24). bits is the minimum
// collsion length.

%{
char *mint(char * today, char* name, int bits)
{
   static byte collision[100];
   char counter[50];
   char *counter_ptr;
   word32 ran0, ran1, ran2, ran3, i0;
   word32 found ;

   memset( counter, '\0', 37 );

   random_start( &ran0, &ran1, &ran2, &ran3 );

   counter_ptr = counter;
   found = 0;

   for ( i0 = 0; !found; )
   {

      sprintf( counter_ptr, GFORMAT GFORMAT, ( i0 + ran0 ) & GROUP_SIZE,
               ran1 & GROUP_SIZE );
      found = find_collision( today, name, bits, collision,
			      GROUP_SIZE, counter_ptr );
      if ( !found )
      {
	 i0 = ( i0 + 1 ) & GROUP_SIZE;
	 if ( i0 == 0 )	/* ie if wrapped */
	 { 
	    return (char *) NULL;
	    /* shouldn't get here without trying  */
	    /* for a very long time */
	 }
      }
   }
   return collision;
}

%}
